@@ -3,30 +3,23 @@
 """
 Gives a multi-value dictionary object (MultiDict) plus several wrappers
 """
-import cgi
-import copy
-import sys
-from webob.util.dictmixin import DictMixin
-try:
-    reversed
-except NameError:
-    from webob.util.reversed import reversed
+import cgi, copy, sys, warnings, urllib
+from UserDict import DictMixin
+
 
 __all__ = ['MultiDict', 'UnicodeMultiDict', 'NestedMultiDict', 'NoVars',
            'TrackableMultiDict']
 
 class MultiDict(DictMixin):
-
     """
-    An ordered dictionary that can have multiple values for each key.
-    Adds the methods getall, getone, mixed, and add to the normal
-    dictionary interface.
+        An ordered dictionary that can have multiple values for each key.
+        Adds the methods getall, getone, mixed and extend and add to the normal
+        dictionary interface.
     """
 
     def __init__(self, *args, **kw):
         if len(args) > 1:
-            raise TypeError(
-                "MultiDict can only be called with one positional argument")
+            raise TypeError("MultiDict can only be called with one positional argument")
         if args:
             if hasattr(args[0], 'iteritems'):
                 items = list(args[0].iteritems())
@@ -40,7 +33,7 @@ class MultiDict(DictMixin):
         if kw:
             self._items.extend(kw.iteritems())
 
-    #@classmethod
+    @classmethod
     def view_list(cls, lst):
         """
         Create a dict that is a view on the given list
@@ -53,25 +46,20 @@ class MultiDict(DictMixin):
         obj._items = lst
         return obj
 
-    view_list = classmethod(view_list)
-
-    #@classmethod
+    @classmethod
     def from_fieldstorage(cls, fs):
         """
         Create a dict from a cgi.FieldStorage instance
         """
         obj = cls()
-        if fs.list:
-            # fs.list can be None when there's nothing to parse
-            for field in fs.list:
-                if field.filename:
-                    obj.add(field.name, field)
-                else:
-                    obj.add(field.name, field.value)
+        # fs.list can be None when there's nothing to parse
+        for field in fs.list or ():
+            if field.filename:
+                obj.add(field.name, field)
+            else:
+                obj.add(field.name, field.value)
         return obj
 
-    from_fieldstorage = classmethod(from_fieldstorage)
-
     def __getitem__(self, key):
         for k, v in reversed(self._items):
             if k == key:
@@ -138,16 +126,12 @@ class MultiDict(DictMixin):
 
     def dict_of_lists(self):
         """
-        Returns a dictionary where each key is associated with a
-        list of values.
+        Returns a dictionary where each key is associated with a list of values.
         """
-        result = {}
-        for key, value in self.iteritems():
-            if key in result:
-                result[key].append(value)
-            else:
-                result[key] = [value]
-        return result
+        r = {}
+        for key, val in self.iteritems():
+            r.setdefault(key, []).append(val)
+        return r
 
     def __delitem__(self, key):
         items = self._items
@@ -197,7 +181,19 @@ class MultiDict(DictMixin):
     def popitem(self):
         return self._items.pop()
 
-    def update(self, other=None, **kwargs):
+    def update(self, *args, **kw):
+        if args:
+            lst = args[0]
+            if len(lst) != len(dict(lst)):
+                # this does not catch the cases where we overwrite existing
+                # keys, but those would produce too many warning
+                msg = ("Behavior of MultiDict.update() has changed "
+                    "and overwrites duplicate keys. Consider using .extend()"
+                )
+                warnings.warn(msg, UserWarning, stacklevel=2)
+        DictMixin.update(self, *args, **kw)
+
+    def extend(self, other=None, **kwargs):
         if other is None:
             pass
         elif hasattr(other, 'items'):
@@ -212,8 +208,8 @@ class MultiDict(DictMixin):
             self.update(kwargs)
 
     def __repr__(self):
-        items = ', '.join(['(%r, %r)' % v for v in self.iteritems()])
-        return '%s([%s])' % (self.__class__.__name__, items)
+        items = map('(%r, %r)'.__mod__, _hide_passwd(self.iteritems()))
+        return '%s([%s])' % (self.__class__.__name__, ', '.join(items))
 
     def __len__(self):
         return len(self._items)
@@ -261,7 +257,7 @@ class UnicodeMultiDict(DictMixin):
     is enabled.
 
     """
-    def __init__(self, multi=None, encoding=None, errors='strict',
+    def __init__(self, multi, encoding=None, errors='strict',
                  decode_keys=False):
         self.multi = multi
         if encoding is None:
@@ -294,10 +290,12 @@ class UnicodeMultiDict(DictMixin):
             # decode FieldStorage's field name and filename
             value = copy.copy(value)
             if self.decode_keys:
-                value.name = value.name.decode(self.encoding, self.errors)
+                if not isinstance(value.name, unicode):
+                    value.name = value.name.decode(self.encoding, self.errors)
             if value.filename:
-                value.filename = value.filename.decode(self.encoding,
-                                                       self.errors)
+                if not isinstance(value.filename, unicode):
+                    value.filename = value.filename.decode(self.encoding,
+                                                           self.errors)
         elif not isinstance(value, unicode):
             try:
                 value = value.decode(self.encoding, self.errors)
@@ -306,7 +304,6 @@ class UnicodeMultiDict(DictMixin):
         return value
 
     def _encode_value(self, value):
-        # FIXME: should this do the FieldStorage stuff too?
         if isinstance(value, unicode):
             value = value.encode(self.encoding, self.errors)
         return value
@@ -327,7 +324,7 @@ class UnicodeMultiDict(DictMixin):
         """
         Return a list of all values matching the key (may be an empty list)
         """
-        return [self._decode_value(v) for v in self.multi.getall(self._encode_key(key))]
+        return map(self._decode_value, self.multi.getall(self._encode_key(key)))
 
     def getone(self, key):
         """
@@ -379,7 +376,9 @@ class UnicodeMultiDict(DictMixin):
         return UnicodeMultiDict(self.multi.copy(), self.encoding, self.errors)
 
     def setdefault(self, key, default=None):
-        return self._decode_value(self.multi.setdefault(self._encode_key(key), self._encode_value(default)))
+        return self._decode_value(
+            self.multi.setdefault(self._encode_key(key),
+                                  self._encode_value(default)))
 
     def pop(self, key, *args):
         return self._decode_value(self.multi.pop(self._encode_key(key), *args))
@@ -389,8 +388,8 @@ class UnicodeMultiDict(DictMixin):
         return (self._decode_key(k), self._decode_value(v))
 
     def __repr__(self):
-        items = ', '.join(['(%r, %r)' % v for v in self.items()])
-        return '%s([%s])' % (self.__class__.__name__, items)
+        items = map('(%r, %r)'.__mod__, _hide_passwd(self.iteritems()))
+        return '%s([%s])' % (self.__class__.__name__, ', '.join(items))
 
     def __len__(self):
         return self.multi.__len__()
@@ -458,18 +457,21 @@ class TrackableMultiDict(MultiDict):
         result = MultiDict.popitem(self)
         self.tracker(self)
         return result
-    def update(self, other=None, **kwargs):
-        MultiDict.update(self, other, **kwargs)
+    def update(self, *args, **kwargs):
+        MultiDict.update(self, *args, **kwargs)
         self.tracker(self)
     def __repr__(self):
-        items = ', '.join(['(%r, %r)' % v for v in self.iteritems()])
-        return '%s([%s])' % (self.name or self.__class__.__name__, items)
+        items = map('(%r, %r)'.__mod__, _hide_passwd(self.iteritems()))
+        return '%s([%s])' % (self.name or self.__class__.__name__, ', '.join(items))
+    def copy(self):
+        # Copies shouldn't be tracked
+        return MultiDict(self)
 
 class NestedMultiDict(MultiDict):
     """
     Wraps several MultiDict objects, treating it as one large MultiDict
     """
-    
+
     def __init__(self, *dicts):
         self.dicts = dicts
 
@@ -550,7 +552,7 @@ class NestedMultiDict(MultiDict):
                 yield key
 
     iterkeys = __iter__
-    
+
 class NoVars(object):
     """
     Represents no variables; used when no variables
@@ -558,7 +560,7 @@ class NoVars(object):
 
     This is read-only
     """
-    
+
     def __init__(self, reason=None):
         self.reason = reason or 'N/A'
 
@@ -618,31 +620,14 @@ class NoVars(object):
     values = keys
     itervalues = iterkeys
 
-__test__ = {
-    'general': """
-    >>> d = MultiDict(a=1, b=2)
-    >>> d['a']
-    1
-    >>> d.getall('c')
-    []
-    >>> d.add('a', 2)
-    >>> d['a']
-    2
-    >>> d.getall('a')
-    [1, 2]
-    >>> d['b'] = 4
-    >>> d.getall('b')
-    [4]
-    >>> d.keys()
-    ['a', 'a', 'b']
-    >>> d.items()
-    [('a', 1), ('a', 2), ('b', 4)]
-    >>> d.mixed() == {'a': [1, 2], 'b': 4}
-    True
-    >>> MultiDict([('a', 'b')], c=2)
-    MultiDict([('a', 'b'), ('c', 2)])
-    """}
 
-if __name__ == '__main__':
-    import doctest
-    doctest.testmod()
+
+def _hide_passwd(items):
+    for k, v in items:
+        if ('password' in k
+            or 'passwd' in k
+            or 'pwd' in k
+        ):
+            yield k, '******'
+        else:
+            yield k, v
